from easy_slam.algorithms.orb_slam import ORBSLAM
from easy_slam.algorithms.fastslam import FastSLAM
from easy_slam.algorithms.graphslam import GraphSLAM
from easy_slam.algorithms.visual_inertial_slam import VisualInertialSLAM
from easy_slam.algorithms.rgbd_slam import RGBDSLAM 